r"""
In practice the following four components are needed to perform psuedo-spectral
projection. (For the "real" spectral projection method, see: :ref:`galerkin`):

-  A distribution for the unknown function parameters (as described in
   section :ref:`distributions`). For example::

      >>> distribution = chaospy.Iid(chaospy.Normal(0, 1), 2)

-  Create integration absissas and weights (as described in :ref:`quadrature`)::

    >>> absissas, weights = chaospy.generate_quadrature(
    ...     2, distribution, rule="G")
    >>> print(numpy.around(absissas, 4))
    [[-1.7321 -1.7321 -1.7321  0.      0.      0.      1.7321  1.7321  1.7321]
     [-1.7321  0.      1.7321 -1.7321  0.      1.7321 -1.7321  0.      1.7321]]
    >>> print(numpy.around(weights, 4))
    [0.0278 0.1111 0.0278 0.1111 0.4444 0.1111 0.0278 0.1111 0.0278]

- An orthogonal polynomial expansion (as described in section
  :ref:`orthogonality`) where the weight function is the distribution in the
  first step::

    >>> expansion = chaospy.orth_ttr(2, distribution)
    >>> print(expansion)
    [1.0, q1, q0, q1^2-1.0, q0q1, q0^2-1.0]

- A function evaluated using the nodes generated in the second step.
  For example::

    >>> def model_solver(q):
    ...     return [q[0]*q[1], q[0]*numpy.e**-q[1]+1]
    >>> solves = [model_solver(absissa) for absissa in absissas.T]
    >>> print(numpy.around(solves[:4], 8))
    [[ 3.         -8.7899559 ]
     [-0.         -0.73205081]
     [-3.          0.69356348]
     [-0.          1.        ]]

- To bring it together, expansion, absissas, weights and solves are used as
  arguments to create approximation::

    >>> approx = chaospy.fit_quadrature(
    ...     expansion, absissas, weights, solves)
    >>> print(chaospy.around(approx, 8))
    [q0q1, -1.58058656q0q1+1.63819248q0+1.0]

Note that in this case the function output is
bivariate. The software is designed to create an approximation of any
discretized model as long as it is compatible with ``numpy`` shapes.

As mentioned in section :ref:`orthogonality`, moment based construction of
polynomials can be unstable. This might also be the case for the
denominator :math:`\mathbb E{\Phi_n^2}`. So when using three terms
recursion, it is common to use the recurrence coefficients to estimated
the denominator.

One cavat with using psuedo-spectral projection is that the calculations of the
norms of the polynomials becomes unstable. To mittigate, recurrence
coefficients can be used to calculate them instead with more stability.
To include these stable norms in the calculations, the following change in code
can be added::

    >>> expansion, norms = chaospy.orth_ttr(2, distribution, retall=True)
    >>> approx2 = chaospy.fit_quadrature(
    ...     expansion, absissas, weights, solves, norms=norms)
    >>> print(chaospy.around(approx2, 8))
    [q0q1, -1.58058656q0q1+1.63819248q0+1.0]

Note that at low polynomial order, the error is very small. For example the
largest coefficient between the two approximation::

    >>> print(numpy.max(abs(approx-approx2).coeffs(), -1) < 1e-12)
    [ True  True]

The ``coeffs`` function returns all the polynomial coefficients.
"""
import numpy
import chaospy


def fit_quadrature(orth, nodes, weights, solves, retall=False, norms=None, **kws):
    """
    Using spectral projection to create a polynomial approximation over
    distribution space.

    Args:
        orth (Poly):
            Orthogonal polynomial expansion. Must be orthogonal for the
            approximation to be accurate.
        nodes (array_like):
            Where to evaluate the polynomial expansion and model to
            approximate.
            nodes.shape==(D,K) where D is the number of dimensions and K is
            the number of nodes.
        weights (array_like):
            Weights when doing numerical integration.
            weights.shape==(K,)
        solves (array_like, callable):
            The model to approximate.
            If array_like is provided, it must have len(solves)==K.
            If callable, it must take a single argument X with len(X)==D,
            and return a consistent numpy compatible shape.
        norms (array_like):
            In the of TTR using coefficients to estimate the polynomial
            norm is more stable than manual calculation.
            Calculated using quadrature if no provided.
            norms.shape==(len(orth),).

    Returns:
        (chaospy.poly.Poly): fitted model approximation.
    """
    orth = chaospy.poly.Poly(orth)
    nodes = numpy.asfarray(nodes)
    weights = numpy.asfarray(weights)

    if callable(solves):
        solves = [solves(node) for node in nodes.T]
    solves = numpy.asfarray(solves)

    shape = solves.shape
    solves = solves.reshape(weights.size, int(solves.size/weights.size))

    ovals = orth(*nodes)
    vals1 = [(val*solves.T*weights).T for val in ovals]

    if norms is None:
        vals2 = [(val**2*weights).T for val in ovals]
        norms = numpy.sum(vals2, 1)
    else:
        norms = numpy.array(norms).flatten()
        assert len(norms) == len(orth)

    coefs = (numpy.sum(vals1, 1).T/norms).T
    coefs = coefs.reshape(len(coefs), *shape[1:])
    approx_model = chaospy.poly.transpose(chaospy.poly.sum(orth*coefs.T, -1))

    if retall:
        return approx_model, coefs
    return approx_model


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Generate samples from `low-discrepancy sequences`_.

In mathematics, a `low-discrepancy sequence`_ is a sequence with the property
that for all values of N, its subsequence x1, ..., xN has a low discrepancy.

Roughly speaking, the discrepancy of a sequence is low if the proportion of
points in the sequence falling into an arbitrary set B is close to proportional
to the measure of B, as would happen on average (but not for particular
samples) in the case of an equidistributed sequence. Specific definitions of
discrepancy differ regarding the choice of B (hyperspheres, hypercubes, etc.)
and how the discrepancy for every B is computed (usually normalized) and
combined (usually by taking the worst value).

Low-discrepancy sequences are also called quasi-random or sub-random sequences,
due to their common use as a replacement of uniformly distributed random
numbers. The "quasi" modifier is used to denote more clearly that the values of
a low-discrepancy sequence are neither random nor pseudorandom, but such
sequences share some properties of random variables and in certain applications
such as the quasi-Monte Carlo method their lower discrepancy is an important
advantage.

.. low-discrepancy sequence: https://en.wikipedia.org/wiki/Low-discrepancy_sequence
"""
from .chebyshev import create_chebyshev_samples, create_nested_chebyshev_samples
from .grid import create_grid_samples, create_nested_grid_samples
from .halton import create_halton_samples
from .hammersley import create_hammersley_samples
from .sobol import create_sobol_samples
from .korobov import create_korobov_samples



"""
Frontend for the collection distributions.

This modules provides a wrapper with documentation for the dist.cores module.
"""
import numpy
from scipy.stats import gaussian_kde

from . import cores, joint


def Alpha(shape=1, scale=1, shift=0):
    """
    Alpha distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scale Parameter
        shift (float, Dist) : Location of lower threshold
    """
    dist = cores.alpha(shape)*scale + shift
    dist.addattr(str="Alpha(%s,%s,%s)" % (shape, scale, shift))
    return dist


def Anglit(loc=0, scale=1):
    """
    Anglit distribution.

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter
    """
    dist = cores.anglit()*scale + loc
    dist.addattr(str="Anglit(%s,%s)"(loc, scale))
    return dist


def Arcsinus(shape=0.5, lo=0, up=1):
    """
    Generalized Arc-sinus distribution

    shape : float, Dist
        Shape parameter where 0.5 is the default non-generalized case.
    lo : float, Dist
        Lower threshold
    up : float, Dist
        Upper threshold
    """
    dist = cores.beta(shape, 1-shape)*(up-lo) + lo
    dist.addattr(str="Arcsinus(%s,%s,%s)" % (shape, lo, up))
    return dist


def Beta(a, b, lo=0, up=1):
    R"""
    Beta Probability Distribution.

    Args:
        a (float, Dist) : First shape parameter, a > 0
        b (float, Dist) : Second shape parameter, b > 0
        lo (float, Dist) : Lower threshold
        up (float, Dist) : Upper threshold

    Examples:
        >>> f = chaospy.Beta(2, 2, 2, 3)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [2.2871 2.4329 2.5671 2.7129]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [2.6039 2.2112 2.8651 2.4881]
        >>> print(f.mom(1))
        2.5
    """
    dist = cores.beta(a, b)*(up-lo) + lo
    dist.addattr(str="Beta(%s,%s,%s,%s)" % (a,b,lo,up))
    return dist


def Bradford(shape=1, lo=0, up=1):
    """
    Bradford distribution.

    Args:
        shape (float, Dist) : Shape parameter
        lo (float, Dist) : Location of lower threshold
        up (float, Dist) : Location of upper threshold
    """
    dist = cores.bradford(c=shape)*(up-lo) + lo
    dist.addattr(str="Bradford(%s,%s,%s)"%(shape, lo, up))
    return dist


def Burr(c=1, d=1, loc=0, scale=1):
    """
    Burr Type III distribution.

    Args:
        c (float, Dist) : Shape parameter
        d (float, Dist) : Shape parameter
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter
    """
    dist = cores.burr(c=c, d=d)*scale + loc
    dist.addattr(str="Burr(%s,%s,%s,%s)"%(c, d, loc, scale))
    return dist


def Cauchy(loc=0, scale=1):
    """
    Cauchy distribution.

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter
    """
    dist = cores.cauchy()*scale + loc
    dist.addattr(str="Cauchy(%s,%s)"%(loc,scale))
    return dist


def Chi(df=1, scale=1, shift=0):
    """
    Chi distribution.

    Args:
        df (float, Dist) : Degrees of freedom
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.chi(df)*scale + shift
    dist.addattr(str="Chi(%s,%s,%s)"%(df, scale, shift))
    return dist


def Chisquard(df=1, scale=1, shift=0, nc=0):
    """
    (Non-central) Chi-squared distribution.

    Args:
        df (float, Dist) : Degrees of freedom
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
        nc (float, Dist) : Non-centrality parameter
    """
    dist = cores.chisquared(df, nc)*scale + shift
    dist.addattr(str="Chisquared(%s,%s,%s,%s)"%(df, nc,scale,shift))
    return dist


def Dbl_gamma(shape=1, scale=1, shift=0):
    """
    Double gamma distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.dbl_gamma(shape)*scale + shift
    dist.addattr(str="Dbl_gamma(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Dbl_weibull(shape=1, scale=1, shift=0):
    """
    Double weibull distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.dbl_weibull(shape)*scale + shift
    dist.addattr(str="Dbl_weibull(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Exponential(scale=1, shift=0):
    R"""
    Exponential Probability Distribution

    Args:
        scale (float, Dist) : Scale parameter. scale!=0
        shift (float, Dist) : Location of the lower bound.

    Examples;:
        >>> f = chaospy.Exponential(1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [0.2231 0.5108 0.9163 1.6094]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [1.0601 0.1222 3.0014 0.6581]
        >>> print(f.mom(1))
        1.0
    """
    dist = cores.expon()*scale + shift
    dist.addattr(str="Expon(%s,%s)" % (scale, shift))
    return dist


def Exponpow(shape=0, scale=1, shift=0):
    """
    Expontial power distribution.

    Also known as Generalized error distribution and Generalized normal
    distribution version 1.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.exponpow(shape)*scale + shift
    dist.addattr(str="Exponpow(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Exponweibull(a=1, c=1, scale=1, shift=0):
    """
    Expontiated Weibull distribution.

    Args:
        a (float, Dist) : First shape parameter
        c (float, Dist) : Second shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.exponweibull(a, c)*scale + shift
    dist.addattr(str="Exponweibull(%s,%s,%s,%s)"%(a, c, scale,shift))
    return dist


def F(n=1, m=1, scale=1, shift=0, nc=0):
    """
    (Non-central) F or Fisher-Snedecor distribution.

    Args:
        n (float, Dist) : Degres of freedom for numerator
        m (float, Dist) : Degres of freedom for denominator
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
        nc (float, Dist) : Non-centrality parameter
    """
    dist = cores.f(n, m, nc)*scale + shift
    dist.addattr(str="F(%s,%s,%s,%s,%s)"%(n, m, scale, shift, nc))
    return dist


def Fatiguelife(shape=1, scale=1, shift=0):
    """
    Fatigue-Life or Birmbaum-Sanders distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.fatiguelife(shape)*scale + shift
    dist.addattr(str="Fatiguelife(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Fisk(shape=1, scale=1, shift=0):
    """
    Fisk or Log-logistic distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.fisk(c=shape)*scale + shift
    dist.addattr(str="Fisk(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Foldcauchy(shape=0, scale=1, shift=0):
    """
    Folded Cauchy distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.foldcauchy(shape)*scale + shift
    dist.addattr(str="Foldcauchy(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Foldnormal(mu=0, sigma=1, loc=0):
    """
    Folded normal distribution.

    Args:
        mu (float, Dist) : Location parameter in normal distribution
        sigma (float, Dist) : Scaling parameter (in both normal and fold)
        loc (float, Dist) : Location of fold
    """
    dist = cores.foldnorm(mu-loc)*sigma + loc
    dist.addattr(str="Foldnorm(%s,%s,%s)"%(mu, sigma, loc))
    return dist


def Frechet(shape=1, scale=1, shift=0):
    """
    Frechet or Extreme value distribution type 2.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.frechet(shape)*scale + shift
    dist.addattr(str="Frechet(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Gamma(shape=1, scale=1, shift=0):
    """
    Gamma distribution.

    Also an Erlang distribution when shape=k and scale=1./lamb.

    Args:
        shape (float, Dist) : Shape parameter. a>0
        scale () : Scale parameter. scale!=0
        shift (float, Dist) : Location of the lower bound.

    Examples:
        >>> f = chaospy.Gamma(1, 1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [0.2231 0.5108 0.9163 1.6094]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [1.0601 0.1222 3.0014 0.6581]
        >>> print(f.mom(1))
        1.0
    """
    dist = cores.gamma(shape)*scale + shift
    dist.addattr(str="Gamma(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Genexpon(a=1, b=1, c=1, scale=1, shift=0):
    """
    Generalized exponential distribution.

    Args:
        a (float, Dist) : First shape parameter
        b (float, Dist) : Second shape parameter
        c (float, Dist) : Third shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter

    Note:
        "An Extension of Marshall and Olkin's Bivariate Exponential Distribution",
        H.K. Ryu, Journal of the American Statistical Association, 1993.

        "The Exponential Distribution: Theory, Methods and Applications",
        N. Balakrishnan, Asit P. Basu.
    """
    dist = cores.genexpon(a=1, b=1, c=1)*scale + shift
    dist.addattr(str="Genexpon(%s,%s,%s)"%(a, b, c))
    return dist


def Genextreme(shape=0, scale=1, loc=0):
    """
    Generalized extreme value distribution
    Fisher-Tippett distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        loc (float, Dist) : Location parameter
    """
    dist = cores.genextreme(shape)*scale + loc
    dist.addattr(str="Genextreme(%s,%s,%s)"%(shape, scale, loc))
    return dist


def Gengamma(shape1, shape2, scale, shift):
    """
    Generalized gamma distribution

    Args:
        shape1 (float, Dist) : Shape parameter 1
        shape2 (float, Dist) : Shape parameter 2
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.gengamma(shape1, shape2)*scale + shift
    dist.addattr(
        str="Gengamma(%s,%s,%s,%s)"%(shape1,shape2,scale,shift))
    return dist


def Genhalflogistic(shape, scale, shift):
    """
    Generalized half-logistic distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.genhalflogistic(shape)*scale + shift
    dist.addattr(str="Genhalflogistic(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Gilbrat(scale=1, shift=0):
    """
    Gilbrat distribution.

    Standard log-normal distribution

    Args:
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.lognormal(1)*scale + shift
    dist.addattr(str="Gilbrat(%s,%s)"%(scale, shift))
    return dist


def Gompertz(shape, scale, shift):
    """
    Gompertz distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.gompertz(shape)*scale + shift
    dist.addattr(str="Gompertz(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Logweibul(scale=1, loc=0):
    """
    Gumbel or Log-Weibull distribution.

    Args:
        scale (float, Dist) : Scaling parameter
        loc (float, Dist) : Location parameter
    """
    dist = cores.gumbel()*scale + loc
    dist.addattr(str="Gumbel(%s,%s)"%(scale, loc))
    return dist


def Hypgeosec(loc=0, scale=1):
    """
    hyperbolic secant distribution

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scale parameter
    """
    dist = cores.hypgeosec()*scale + loc
    dist.addattr(str="Hypgeosec(%s,%s)"%(loc, scale))
    return dist


def Kumaraswamy(a, b, lo=0, up=1):
    """
    Kumaraswswamy's double bounded distribution

    Args:
        a (float, Dist) : First shape parameter
        b (float, Dist) : Second shape parameter
        lo (float, Dist) : Lower threshold
        up (float, Dist) : Upper threshold
    """
    dist = cores.kumaraswamy(a,b)*(up-lo) + lo
    dist.addattr(str="Kumaraswamy(%s,%s,%s,%s)"%(a,b,lo,up))
    return dist


def Laplace(mu=0, scale=1):
    R"""
    Laplace Probability Distribution

    Args:
        mu (float, Dist) : Mean of the distribution.
        scale (float, Dist) : Scaleing parameter.
            scale > 0

    Examples:
        >>> f = chaospy.Laplace(2, 2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [0.1674 1.5537 2.4463 3.8326]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [ 2.734  -0.9392  6.6165  1.9275]
        >>> print(f.mom(1))
        2.0
    """
    dist = cores.laplace()*scale + mu
    dist.addattr(str="Laplace(%s,%s)"%(mu,scale))
    return dist


def Levy(loc=0, scale=1):
    """
    Levy distribution

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter
    """
    dist = cores.levy()*scale+loc
    dist.addattr(str="Levy(%s,%s)"%(loc, scale))
    return dist


def Loggamma(shape=1, scale=1, shift=0):
    """
    Log-gamma distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.loggamma(shape)*scale + shift
    dist.addattr(str="Loggamma(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Logistic(loc=0, scale=1, skew=1):
    """
    Generalized logistic type 1 distribution
    Sech squared distribution

    loc (float, Dist) : Location parameter
    scale (float, Dist) : Scale parameter
    skew (float, Dist) : Shape parameter
    """
    dist = cores.logistic()*scale + loc
    dist.addattr(str="Logistic(%s,%s)"%(loc, scale))
    return dist


def Loglaplace(shape=1, scale=1, shift=0):
    """
    Log-laplace distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.loglaplace(shape)*scale + shift
    dist.addattr(str="Loglaplace(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Lognormal(mu=0, sigma=1, shift=0, scale=1):
    R"""
    Log-normal distribution

    Args:
        mu (float, Dist) : Mean in the normal distribution.  Overlaps with
                scale by mu=log(scale)
        sigma (float, Dist) : Standard deviation of the normal distribution.
        shift (float, Dist) : Location of the lower bound.
        scale (float, Dist) : Scale parameter. Overlaps with mu by scale=e**mu

    Examples:
        >>> f = chaospy.Lognormal(0, 1)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [0.431  0.7762 1.2883 2.3201]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [1.4844 0.3011 5.1945 0.9563]
        >>> print(numpy.around(f.mom(1), 4))
        1.6487
    """
    dist = cores.lognormal(sigma)*scale*numpy.e**mu + shift
    dist.addattr(str="Lognormal(%s,%s,%s,%s)"%(mu,sigma,shift,scale))
    return dist


def Loguniform(lo=0, up=1, scale=1, shift=0):
    """
    Log-uniform distribution

    Args:
        lo (float, Dist) : Location of lower threshold
        up (float, Dist) : Location of upper threshold
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.loguniform(lo, up)*scale + shift
    dist.addattr(str="Loguniform(%s,%s,%s,%s)" % (lo,up,scale,shift))
    return dist


def Maxwell(scale=1, shift=0):
    """
    Maxwell-Boltzmann distribution
    Chi distribution with 3 degrees of freedom

    Args:
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.chi(3)*scale + shift
    dist.addattr(str="Maxwell(%s,%s)"%(scale, shift))
    return dist


def Mielke(kappa=1, expo=1, scale=1, shift=0):
    """
    Mielke's beta-kappa distribution

    Args:
        kappa (float, Dist) : First shape parameter
        expo (float, Dist) : Second shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.mielke(kappa, expo)*scale + shift
    dist.addattr(str="Mielke(%s,%s,%s,%s)"%(kappa,expo,scale,shift))
    return dist


def MvLognormal(loc=[0,0], scale=[[1,.5],[.5,1]]):
    """
    Multivariate Log-Normal Distribution.

    Args:
        loc (float, Dist) : Mean vector
        scale (float, Dist) : Covariance matrix or variance vector if scale is
                a 1-d vector.
    """
    dist = cores.mvlognormal(loc, scale)
    dist.addattr(str="MvLognormal(%s,%s)" % (loc, scale))
    return dist


def MvNormal(loc=[0,0], scale=[[1,.5],[.5,1]]):
    """
    Multivariate Normal Distribution

    Args:
        loc (float, Dist) : Mean vector
        scale (float, Dist) : Covariance matrix or variance vector if scale is a 1-d vector.

    Examples:
        >>> f = chaospy.MvNormal([0,0], [[1,.5],[.5,1]])
        >>> q = [[.4,.5,.6],[.4,.5,.6]]
        >>> print(numpy.around(f.inv(q), 4))
        [[-0.2533  0.      0.2533]
         [-0.3461  0.      0.3461]]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [[0.4 0.5 0.6]
         [0.4 0.5 0.6]]
        >>> print(numpy.around(f.sample(3), 4))
        [[ 0.395  -1.2003  1.6476]
         [ 0.1588  0.3855  0.1324]]
        >>> print(numpy.around(f.mom((1,1)), 4))
        0.5
    """
    if numpy.all((numpy.diag(numpy.diag(scale))-scale)==0):
        out = joint.J(
            *[Normal(loc[i], scale[i,i]) for i in range(len(scale))])
    else:
        out = cores.mvnormal(loc, scale)
    out.addattr(str="MvNormal(%s,%s)" % (loc, scale))
    return out


def MvStudent_t(df=1, loc=[0,0], scale=[[1,.5],[.5,1]]):
    """
    Args:
        df (float, Dist) : Degree of freedom
        loc (array_like, Dist) : Location parameter
        scale (array_like) : Covariance matrix
    """
    out = cores.mvstudentt(df, loc, scale)
    out.addattr(str="MvStudent_t(%s,%s,%s)" % (df, loc, scale))
    return out


def Nakagami(shape=1, scale=1, shift=0):
    """
    Nakagami-m distribution.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.nakagami(shape)*scale + shift
    dist.addattr(str="Nakagami(%s,%s,%s)"%(shape,scale,shift))
    return dist


def Normal(mu=0, sigma=1):
    R"""
    Normal (Gaussian) distribution

    Args:
        mu (float, Dist) : Mean of the distribution.
        sigma (float, Dist) : Standard deviation.  sigma > 0

    Examples:
        >>> f = chaospy.Normal(2, 2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [0.3168 1.4933 2.5067 3.6832]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [ 2.7901 -0.4006  5.2952  1.9107]
        >>> print(f.mom(1))
        2.0
    """
    dist = cores.normal()*sigma + mu
    dist.addattr(str="Normal(%s,%s)"%(mu, sigma))
    return dist


def OTDistribution(distribution):
    """
    OpenTURNS distribution.

    Args:
        distribution (openturns.Distribution, Dist) : underlying OT distribution
    """
    dist = cores.otdistribution(distribution)
    return dist


def Pareto1(shape=1, scale=1, loc=0):
    """
    Pareto type 1 distribution.

    Lower threshold at scale+loc and survival: x^-shape

    Args:
        shape (float, Dist) : Tail index parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.pareto1(shape)*scale + loc
    dist.addattr(str="Pareto(%s,%s,%s)" % (shape, scale, loc))
    return dist


def Pareto2(shape=1, scale=1, loc=0):
    """
    Pareto type 2 distribution.

    Also known as Lomax distribution (for loc=0).

    Lower threshold at loc and survival: (1+x)^-shape.

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        loc (float, Dist) : Location parameter
    """
    dist = cores.pareto(shape)*scale + loc
    dist.addattr(str="Pareto(%s,%s,%s)"%(shape, scale, loc))
    return dist


def Powerlaw(shape=1, lo=0, up=1):
    """
    Powerlaw distribution

    Args:
        shape (float, Dist) : Shape parameter
        lo (float, Dist) : Location of lower threshold
        up (float, Dist) : Location of upper threshold
    """
    dist = cores.beta(shape, 1)*(up-lo) + lo
    dist.addattr(str="Powerlaw(%s,%s,%s)"%(shape, lo, up))
    return dist


def Powerlognormal(shape=1, mu=0, sigma=1, shift=0, scale=1):
    """
    Power log-normal distribution

    Args:
        shape (float, Dist) : Shape parameter
        mu (float, Dist) : Mean in the normal distribution.  Overlaps with
                scale by mu=log(scale)
        sigma (float, Dist) : Standard deviation of the normal distribution.
        shift (float, Dist) : Location parameter
        scale (float, Dist) : Scaling parameter. Overlap with mu in scale=e**mu
    """
    dist = cores.powerlognorm(shape, sigma)*scale*numpy.e**mu + shift
    dist.addattr(str="Powerlognorm(%s,%s,%s,%s,%s)"%\
            (shape, mu, sigma, shift, scale))
    return dist


def Powernorm(shape=1, mu=0, scale=1):
    """
    Power normal or Box-Cox distribution.

    Args:
        shape (float, Dist) : Shape parameter
        mu (float, Dist) : Mean of the normal distribution
        scale (float, Dist) : Standard deviation of the normal distribution
    """
    dist = cores.powernorm(shape)*scale + mu
    dist.addattr(str="Powernorm(%s,%s,%s)"%(shape, mu, scale))
    return dist


def Raised_cosine(loc=0, scale=1):
    """
    Raised cosine distribution

    Args:
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scale parameter
    """
    dist = cores.raised_cosine()*scale + loc
    dist.addattr(str="Raised_cosine(%s,%s)"%(loc,scale))
    return dist


def Rayleigh(scale=1, shift=0):
    """
    Rayleigh distribution

    Args:
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.chi(2)*scale + shift
    dist.addattr(str="Rayleigh(%s,%s)"%(scale, shift))
    return dist


def Reciprocal(lo=1, up=2):
    """
    Reciprocal distribution

    Args:
        lo (float, Dist) : Location of lower threshold
        up (float, Dist) : Location of upper threshold
    """
    dist = cores.reciprocal(lo,up)
    dist.addattr(str="Reciprocal(%s,%s)"%(lo,up))
    return dist


def Student_t(df, loc=0, scale=1, nc=0):
    """
    (Non-central) Student-t distribution

    Args:
        df (float, Dist) : Degrees of freedom
        loc (float, Dist) : Location parameter
        scale (float, Dist) : Scale parameter
        nc (flat, Dist) : Non-centrality parameter
    """
    dist = cores.student_t(df)*scale + loc
    dist.addattr(str="Student_t(%s,%s,%s)" % (df, loc, scale))
    return dist


def Triangle(lo, mid, up):
    """
    Triangle Distribution.

    Must have lo <= mid <= up.

    Args:
        lo (float, Dist) : Lower bound
        mid (float, Dist) : Location of the top
        up (float, Dist) : Upper bound

    Examples:
        >>> f = chaospy.Triangle(2, 3, 4)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [2.6325 2.8944 3.1056 3.3675]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [3.1676 2.4796 3.6847 2.982 ]
        >>> print(numpy.around(f.mom(1), 4))
        3.0
    """
    dist = cores.triangle((mid-lo)*1./(up-lo))*(up-lo) + lo
    dist.addattr(str="Triangle(%s,%s,%s)" % (lo, mid, up))
    return dist


def Truncexpon(up=1, scale=1, shift=0):
    """
    Truncated exponential distribution.

    Args:
        up (float, Dist) : Location of upper threshold
        scale (float, Dist) : Scaling parameter in the exponential distribution
        shift (float, Dist) : Location parameter
    """
    dist = cores.truncexpon((up-shift)/scale)*scale + shift
    dist.addattr(str="Truncexpon(%s,%s,%s)"%(up, scale, shift))
    return dist


def Truncnorm(lo=-1, up=1, mu=0, sigma=1):
    """
    Truncated normal distribution

    Args:
        lo (float, Dist) : Location of lower threshold
        up (float, Dist) : Location of upper threshold
        mu (float, Dist) : Mean of normal distribution
        sigma (float, Dist) : Standard deviation of normal distribution
    """
    dist = cores.truncnorm(lo, up, mu, sigma)
    dist.addattr(str="Truncnorm(%s,%s,%s,%s)"%(lo,up,mu,sigma))
    return dist


def Tukeylambda(shape=0, scale=1, shift=0):
    """
    Tukey-lambda distribution

    Args:
        lam (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.tukeylambda(shape)*scale + shift
    dist.addattr(str="Tukeylambda(%s,%s,%s)"%(shape, scale, shift))
    return dist


def Uniform(lo=0, up=1):
    r"""
    Uniform distribution

    Args:
        lo (float, Dist) : Lower threshold of distribution. Must be smaller than up.
        up (float, Dist) : Upper threshold of distribution.

    Examples:
        >>> f = chaospy.Uniform(2, 4)
        >>> q = numpy.linspace(0,1,5)
        >>> print(numpy.around(f.inv(q), 4))
        [2.  2.5 3.  3.5 4. ]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.   0.25 0.5  0.75 1.  ]
        >>> print(numpy.around(f.sample(4), 4))
        [3.3072 2.23   3.9006 2.9644]
        >>> print(f.mom(1))
        3.0
    """

    dist = cores.uniform()*((up-lo)*.5)+((up+lo)*.5)
    dist.addattr(str="Uniform(%s,%s)"%(lo,up))
    return dist


def Wald(mu=0, scale=1, shift=0):
    """
    Wald distribution
    Reciprocal inverse Gaussian distribution

    Args:
        mu (float, Dist) : Mean of the normal distribution
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.wald(mu)*scale + shift
    dist.addattr(str="Wald(%s,%s,%s)"%(mu, scale, shift))
    return dist


def Weibull(shape=1, scale=1, shift=0):
    """
    Weibull Distribution

    Args:
        shape (float, Dist) : Shape parameter.
        scale (float, Dist) : Scale parameter.
        shift (float, Dist) : Location of lower bound.

    Examples:
        >>> f = chaospy.Weibull(2)
        >>> q = numpy.linspace(0,1,6)[1:-1]
        >>> print(numpy.around(f.inv(q), 4))
        [0.4724 0.7147 0.9572 1.2686]
        >>> print(numpy.around(f.fwd(f.inv(q)), 4))
        [0.2 0.4 0.6 0.8]
        >>> print(numpy.around(f.sample(4), 4))
        [1.0296 0.3495 1.7325 0.8113]
        >>> print(numpy.around(f.mom(1), 4))
        0.8862
    """
    dist = cores.weibull(shape)*scale + shift
    dist.addattr(str="Weibull(%s,%s,%s)" % (shape, scale, shift))
    return dist


def Wigner(radius=1, shift=0):
    """
    Wigner (semi-circle) distribution

    Args:
        radius (float, Dist) : radius of the semi-circle (scale)
        shift (float, Dist) : location of the origen (location)
    """
    dist = radius*(2*cores.beta(1.5,1.5)-1) + shift
    dist.addattr(str="Wigner(%s,%s)" % (radius, shift))
    return dist


def Wrapcauchy(shape=0.5, scale=1, shift=0):
    """
    Wraped Cauchy distribution

    Args:
        shape (float, Dist) : Shape parameter
        scale (float, Dist) : Scaling parameter
        shift (float, Dist) : Location parameter
    """
    dist = cores.wrapcauchy(shape)*scale + shift
    dist.addattr(str="Wrapcauchy(%s,%s,%s)"%(shape, scale, shift))
    return dist


def SampleDist(samples, lo=None, up=None):
    """
    Distribution based on samples.

    Estimates a distribution from the given samples by constructing a kernel
    density estimator (KDE).

    Args:
        samples:
            Sample values to construction of the KDE
        lo (float) : Location of lower threshold
        up (float) : Location of upper threshold
    """
    if lo is None:
        lo = samples.min()
    if up is None:
        up = samples.max()

    try:
        #construct the kernel density estimator
        kernel = gaussian_kde(samples, bw_method="scott")
        dist = cores.kdedist(kernel, lo, up)
        dist.addattr(str="SampleDist(%s,%s)" % (lo, up))

    #raised by gaussian_kde if dataset is singular matrix
    except numpy.linalg.LinAlgError:
        dist = Uniform(lo=-numpy.inf, up=numpy.inf)

    return dist

if __name__=='__main__':
    import __init__ as cp
    import numpy as np
    import doctest
    doctest.testmod()




"""
Create `antithetic variates`_ from variables on the unit hyper-cube.

In statistics, the antithetic variates method is a variance reduction technique
used in Monte Carlo methods. Considering that the error reduction in the
simulated signal (using Monte Carlo methods) has a square root convergence,
a very large number of sample paths is required to obtain an accurate result.
The antithetic variates method reduces the variance of the simulation results.

Example usage
-------------

Univariate case::

    >>> print(create_antithetic_variates([0.1, 0.11, 0.88]))
    [[0.1  0.9  0.11 0.89 0.88 0.12]]

Bivariate case::

    >>> print(create_antithetic_variates([[0.11, 0.22], [0.55, 0.66]]))
    [[0.11 0.89 0.11 0.89 0.22 0.78 0.22 0.78]
     [0.55 0.55 0.45 0.45 0.66 0.66 0.34 0.34]]

Freeze axes::

    >>> print(create_antithetic_variates([[0.1, 0.11], [0.88, 0.99]], axes=[1, 0]))
    [[0.1  0.9  0.11 0.89]
     [0.88 0.88 0.99 0.99]]
    >>> print(create_antithetic_variates([[0.1, 0.11], [0.88, 0.99]], axes=[0, 1]))
    [[0.1  0.1  0.11 0.11]
     [0.88 0.12 0.99 0.01]]

Multivariate case::

    >>> print(create_antithetic_variates(
    ...     [[0.1, 0.11], [0.2, 0.22], [0.3, 0.33]], axes=[1, 0, 0]))
    [[0.1  0.9  0.11 0.89]
     [0.2  0.2  0.22 0.22]
     [0.3  0.3  0.33 0.33]]
    >>> print(create_antithetic_variates(
    ...     [[0.1, 0.11], [0.2, 0.22], [0.3, 0.33]], axes=[0, 1, 0]))
    [[0.1  0.1  0.11 0.11]
     [0.2  0.8  0.22 0.78]
     [0.3  0.3  0.33 0.33]]
    >>> print(create_antithetic_variates(
    ...     [[0.1, 0.11], [0.2, 0.22], [0.3, 0.33]], axes=[0, 0, 1]))
    [[0.1  0.1  0.11 0.11]
     [0.2  0.2  0.22 0.22]
     [0.3  0.7  0.33 0.67]]

.. _antithetic variates: https://en.wikipedia.org/wiki/Antithetic_variates>
"""
import numpy


def create_antithetic_variates(samples, axes=None):
    """
    Generate antithetic variables.

    Args:
        samples (array_like):
            The samples, assumed to be on the [0, 1]^D hyper-cube, to be
            reflected.
        axes (array_like, optional):
            Boolean array of which axes to reflect. If This to limit the number
            of points created in higher dimensions by reflecting all axes at
            once.

    Returns (numpy.ndarray):
        Same as ``samples``, but with samples internally reflected. roughly
        equivalent to ``numpy.vstack([samples, 1-samples])`` in one dimensions.
    """
    samples = numpy.asfarray(samples)
    assert numpy.all(samples <= 1) and numpy.all(samples >= 0), (
        "all samples assumed on interval [0, 1].")
    if len(samples.shape) == 1:
        samples = samples.reshape(1, -1)
    inverse_samples = 1-samples
    dims = len(samples)

    if axes is None:
        axes = True
    axes = numpy.array(axes, dtype=bool).flatten()

    indices = {tuple(axes*idx) for idx in numpy.ndindex((2,)*dims)}
    indices = sorted(indices, reverse=True)
    indices = sorted(indices, key=lambda idx: sum(idx))
    out = [numpy.where(idx, inverse_samples.T, samples.T).T for idx in indices]
    out = numpy.dstack(out).reshape(dims, -1)
    return out



# pylint: disable=protected-access
"""Backend for three terms recursion generator."""
import numpy


def ttr_call(self, order, dist):
    "TTR call backend wrapper"

    assert order.shape == (len(dist), self.size)
    graph = self.graph
    self.dist, dist_ = dist, self.dist

    graph.add_node(dist, key=order)
    if hasattr(dist, "_ttr"):
        if dist.advance:
            out = dist._ttr(order, self)
        else:
            out = numpy.empty((2,)+order.shape)
            prm = self.dists.build()
            prm.update(self.keys.build())
            out[0], out[1] = dist._ttr(order, **prm)
    else:
        raise NotImplementedError(
            "No `_ttr` method found for %s" % dist)

    graph.add_node(dist, val=out)

    self.dist = dist_
    return numpy.array(out)



import numpy as np

import chaospy

from .second1d import Var
from .first import E_cond
from .first import E as E_total

def Sens_m(poly, dist, **kws):
    """
    Variance-based decomposition
    AKA Sobol' indices

    First order sensitivity indices
    """

    dim = len(dist)
    if poly.dim<dim:
        poly = chaospy.poly.setdim(poly, len(dist))

    zero = [0]*dim
    out = np.zeros((dim,) + poly.shape)
    V = Var(poly, dist, **kws)
    for i in range(dim):
        zero[i] = 1
        out[i] = Var(E_cond(poly, zero, dist, **kws),
                     dist, **kws)/(V+(V == 0))*(V != 0)
        zero[i] = 0
    return out


def Sens_m2(poly, dist, **kws):
    """
    Variance-based decomposition
    AKA Sobol' indices

    Second order sensitivity indices
    """
    dim = len(dist)
    if poly.dim<dim:
        poly = chaospy.poly.setdim(poly, len(dist))

    zero = [0]*dim
    out = np.zeros((dim, dim) + poly.shape)

    mean = E_total(poly, dist)
    V_total = Var(poly, dist)
    E_cond_i = [None]*dim
    V_E_cond_i = [None]*dim
    for i in range(dim):
        zero[i] = 1
        E_cond_i[i] = E_cond(poly, zero, dist, **kws) 
        V_E_cond_i[i] = Var(E_cond_i[i], dist, **kws)
        zero[i] = 0
             
    for i in range(dim):
        zero[i] = 1
        for j in range(i+1, dim):
            zero[j] = 1
            E_cond_ij = E_cond(poly, zero, dist, **kws)
            out[j, i] = out[i, j] = (Var(E_cond_ij, dist, **kws)-V_E_cond_i[i] - V_E_cond_i[j]) /(V_total+(V_total == 0))*(V_total != 0)
            zero[j] = 0

        zero[i] = 0

    return out


def Sens_t(poly, dist, **kws):
    """
    Variance-based decomposition
    AKA Sobol' indices

    Total effect sensitivity index
    """
    dim = len(dist)
    if poly.dim<dim:
        poly = chaospy.poly.setdim(poly, len(dist))

    zero = [1]*dim
    out = np.zeros((dim,) + poly.shape, dtype=float)
    V = Var(poly, dist, **kws)
    for i in range(dim):
        zero[i] = 0
        out[i] = (V-Var(E_cond(poly, zero, dist, **kws),
            dist, **kws))/(V+(V==0))**(V!=0)
        zero[i] = 1
    return out


def Sens_m_nataf(order, dist, samples, vals, **kws):
    """
    Variance-based decomposition thorugh the Nataf distribution.

    Generates first order sensitivity indices

    Args:
        order (int): polynomial order used `orth_ttr`.
        dist (Copula): Assumed to be Nataf with independent components
        samples (array_like): Samples used for evaluation (typically generated
                from `dist`.)
        vals (array_like): Evaluations of the model for given samples.

    Returns:
        np.ndarray: Sensitivity indices with
                `shape==(len(dist),) + vals.shape[1:]`
    """
    assert dist.__class__.__name__ == "Copula"
    trans = dist.prm["trans"]
    assert trans.__class__.__name__ == "nataf"
    vals = np.array(vals)

    cov = trans.prm["C"]
    cov = np.dot(cov, cov.T)

    marginal = dist.prm["dist"]
    dim = len(dist)

    orth = chaospy.orthogonal.orth_ttr(order, marginal, sort="GR")

    r = range(dim)

    index = [1] + [0]*(dim-1)

    nataf = chaospy.dist.Nataf(marginal, cov, r)
    samples_ = marginal.inv( nataf.fwd( samples ) )
    poly, coeffs = chaospy.collocation.fit_regression(
        orth, samples_, vals, retall=1)

    V = Var(poly, marginal, **kws)

    out = np.zeros((dim,) + poly.shape)
    out[0] = Var(E_cond(poly, index, marginal, **kws),
                 marginal, **kws)/(V+(V == 0))*(V != 0)


    for i in range(1, dim):

        r = r[1:] + r[:1]
        index = index[-1:] + index[:-1]

        nataf = chaospy.dist.Nataf(marginal, cov, r)
        samples_ = marginal.inv( nataf.fwd( samples ) )
        poly, coeffs = chaospy.collocation.fit_regression(
            orth, samples_, vals, retall=1)

        out[i] = Var(E_cond(poly, index, marginal, **kws),
                     marginal, **kws)/(V+(V == 0))*(V != 0)

    return out


def Sens_t_nataf(order, dist, samples, vals, **kws):
    """
    Variance-based decomposition thorugh the Nataf distribution.

    Total order sensitivity indices

    Args:
        order (int): polynomial order used `orth_ttr`.
        dist (Copula): Assumed to be Nataf with independent components
        samples (array_like): Samples used for evaluation (typically generated
                from `dist`.)
        vals (array_like): Evaluations of the model for given samples.

    Returns:
        np.ndarray: Sensitivity indices with
                `shape==(len(dist),)+vals.shape[1:]`
    """

    assert dist.__class__.__name__ == "Copula"
    trans = dist.prm["trans"]
    assert trans.__class__.__name__ == "nataf"
    vals = np.array(vals)

    cov = trans.prm["C"]
    cov = np.dot(cov, cov.T)

    marginal = dist.prm["dist"]
    dim = len(dist)

    orth = chaospy.orthogonal.orth_ttr(order, marginal, sort="GR")

    r = range(dim)

    index = [0] + [1]*(dim-1)

    nataf = chaospy.dist.Nataf(marginal, cov, r)
    samples_ = marginal.inv( nataf.fwd( samples ) )
    poly, coeffs = chaospy.collocation.fit_regression(
        orth, samples_, vals, retall=1)

    V = Var(poly, marginal, **kws)

    out = np.zeros((dim,) + poly.shape)
    out[0] = (V-Var(E_cond(poly, index, marginal, **kws),
                    marginal, **kws))/(V+(V == 0))**(V != 0)

    for i in range(1, dim):

        r = r[1:] + r[:1]
        index = index[-1:] + index[:-1]

        nataf = chaospy.dist.Nataf(marginal, cov, r)
        samples_ = marginal.inv( nataf.fwd( samples ) )
        poly, coeffs = chaospy.collocation.fit_regression(
            orth, samples_, vals, retall=1)

        out[i] = (V-Var(E_cond(poly, index, marginal, **kws),
                        marginal, **kws))/(V+(V == 0))*(V != 0)

    return out


def Sens_nataf(order, dist, samples, vals, **kws):
    """
    Variance-based decomposition thorugh the Nataf distribution.

    Main and total order sensitivity indices

    Args:
        order (int): polynomial order used `orth_ttr`.
        dist (Copula): Assumed to be Nataf with independent components
        samples (array_like): Samples used for evaluation (typically generated
                from `dist`.)
        vals (array_like): Evaluations of the model for given samples.

    Returns:
        np.ndarray: Sensitivity indices with
                `shape==(2, len(dist),)+vals.shape[1:]`. First component is
                main and second is total.
    """

    assert dist.__class__.__name__ == "Copula"
    trans = dist.prm["trans"]
    assert trans.__class__.__name__ == "nataf"
    vals = np.array(vals)

    cov = trans.prm["C"]
    cov = np.dot(cov, cov.T)

    marginal = dist.prm["dist"]
    dim = len(dist)

    orth = chaospy.orthogonal.orth_ttr(order, marginal, sort="GR")

    r = range(dim)

    index0 = [0] + [1]*(dim-1)
    index1 = [1] + [0]*(dim-1)

    nataf = chaospy.dist.Nataf(marginal, cov, r)
    samples_ = marginal.inv( nataf.fwd( samples ) )
    poly, coeffs = chaospy.collocation.fit_regression(
        orth, samples_, vals, retall=1)

    V = Var(poly, marginal, **kws)

    out = np.zeros((2, dim,) + poly.shape)
    out[0, 0] = (V - Var(E_cond(poly, index0, marginal, **kws),
                        marginal, **kws))/(V+(V == 0))**(V != 0)
    out[1, 0] = Var(E_cond(poly, index1, marginal, **kws),
                   marginal, **kws)/(V+(V == 0))*(V != 0)

    for i in range(1, dim):

        r = r[1:] + r[:1]
        index0 = index0[-1:] + index0[:-1]

        nataf = chaospy.dist.Nataf(marginal, cov, r)
        samples_ = marginal.inv( nataf.fwd( samples ) )
        poly, coeffs = chaospy.collocation.fit_regression(
            orth, samples_, vals, retall=1)

        out[0, i] = (V-Var(E_cond(poly, index0, marginal, **kws),
                          marginal, **kws))/(V+(V == 0))*(V != 0)
        out[1, i] = Var(E_cond(poly, index1, marginal, **kws),
                       marginal, **kws)/(V+(V == 0))*(V != 0)

    return out[::-1]



import numpy
from scipy import special

from .baseclass import Archimedean, Copula
from ..baseclass import Dist


class gumbel(Archimedean):
    "Gumbel copula backend"

    def __init__(self, N, theta=1., eps=1e-6):
        theta = float(theta)
        Dist.__init__(self, th=theta, eps=eps, _length=N)
    def gen(self, x, th):
        return (-numpy.log(x))**th
    def igen(self, x, th):
        return numpy.e**(-x**th)


def Gumbel(dist, theta=2., eps=1e-6):
    r"""
    Gumbel Copula

    .. math::
        \phi(x;th) = \frac{x^{-th}-1}{th}
        \phi^{-1}(q;th) = (q*th + 1)^{-1/th}

    where `th` (or theta) is defined on the interval `[1,inf)`.

Args:
    dist (Dist) : The Distribution to wrap
    theta (float) : Copula parameter

Returns:
    (Dist) : The resulting copula distribution.

Examples:
    >>> dist = chaospy.J(chaospy.Uniform(), chaospy.Normal())
    >>> copula = chaospy.Gumbel(dist, theta=2)
    >>> print(copula.sample(3, "H"))
    [[ 0.125       0.625       0.375     ]
     [ 0.62638094 -0.45417192 -0.21628863]]
"""
    return Copula(dist, gumbel(len(dist), theta, eps))


class clayton(Archimedean):
    "clayton copula backend"

    def __init__(self, N, theta=1., eps=1e-6):
        Dist.__init__(self, th=float(theta), _length=N, eps=eps)
    def gen(self, x, th):
        return (x**-th-1)/th
    def igen(self, x, th):
        return (1.+th*x)**(-1./th)

def Clayton(dist, theta=2., eps=1e-6):
    return Copula(dist, clayton(len(dist), theta, eps))


class ali_mikhail_haq(Archimedean):
    "Ali Mikhail Haq copula backend"

    def __init__(self, N, theta=.5, eps=1e-6):
        theta = float(theta)
        assert -1<=theta<1
        Dist.__init__(self, th=theta, _length=N, eps=eps)
    def gen(self, x, th):
        return numpy.log((1-th*(1-x))/x)
    def igen(self, x, th):
        return (1-th)/(numpy.e**x-th)


def Ali_mikhail_haq(dist, theta=2., eps=1e-6):
    "Ali Mikhail Haq copula"
    trans = ali_mikhail_haq(len(dist), theta, eps)
    return Copula(dist, trans)

class frank(Archimedean):
    "Frank copula backend"

    def __init__(self, N, theta, eps=1e-6):
        "theta!=0"
        theta = float(theta)
        assert theta!=0
        Dist.__init__(self, th=theta, _length=N, eps=eps)

    def gen(self, x, th):
        return -numpy.log((numpy.e**(-th*x)-1)/(numpy.e**-th-1))
    def igen(self, q, th):
        return -numpy.log(1+numpy.e**-q*(numpy.e**-th-1))/th

def Frank(dist, theta=1., eps=1e-4):
    "Frank copula"
    return Copula(dist, frank(len(dist), theta, eps))

class joe(Archimedean):
    "Joe copula backend"

    def __init__(self, N, theta, eps=1e-6):
        "theta in [1,inf)"
        theta = float(theta)
        assert theta>=1
        Dist.__init__(self, th=theta, _length=N, eps=eps)

    def gen(self, x, th):
        return -numpy.log(1-(1-x)**th)

    def igen(self, q, th):
        return 1-(1-numpy.e**-q)**(1/th)

def Joe(dist, theta=2., eps=1e-6):
    "Joe copula"
    return Copula(dist, joe(len(dist), theta, eps))

class nataf(Dist):
    "Nataf (normal) copula backend"

    def __init__(self, R, ordering=None):
        "R symmetric & positive definite matrix"

        if ordering is None:
            ordering = range(len(R))
        ordering = numpy.array(ordering)

        P = numpy.eye(len(R))[ordering]

        R = numpy.dot(P, numpy.dot(R, P.T))
        R = numpy.linalg.cholesky(R)
        R = numpy.dot(P.T, numpy.dot(R, P))
        Ci = numpy.linalg.inv(R)
        Dist.__init__(self, C=R, Ci=Ci, _length=len(R))

    def _cdf(self, x, C, Ci):
        out = special.ndtr(numpy.dot(Ci, special.ndtri(x)))
        return out

    def _ppf(self, q, C, Ci):
        out = special.ndtr(numpy.dot(C, special.ndtri(q)))
        return out

    def _bnd(self, C, Ci):
        return 0.,1.

def Nataf(dist, R, ordering=None):
    "Nataf (normal) copula"
    return Copula(dist, nataf(R, ordering))


class t_copula(Dist):

    def __init__(self, a, R):
        from ..cores import mvstudentt, student_t
        self.MV = mvstudentt(a, numpy.zeros(len(R)), R)
        self.UV = student_t(a)
        Dist.__init__(self, _length=len(R))

    def _cdf(self, x):
        out = self.MV.fwd(self.UV.inv(x))
        return out

    def _ppf(self, q):
        out = self.MV.inv(q)
        out = self.UV.fwd(out)
        return out

    def _bnd(self):
        return 0.,1.

def T_copula(dist, a, R):
    return Copula(dist, t_copula(a, R))



"""
Functions that performes arithmetics on polynomial.
"""

import numpy as np

import chaospy.poly.dimension
import chaospy.poly.typing

from chaospy.poly.base import Poly


def add(*args):
    """Polynomial addition."""
    if len(args) > 2:
        return add(args[0], add(args[1], args[1:]))

    if len(args) == 1:
        return args[0]

    part1, part2 = args

    if isinstance(part2, Poly):

        if part2.dim > part1.dim:
            part1 = chaospy.dimension.setdim(part1, part2.dim)
        elif part2.dim < part1.dim:
            part2 = chaospy.dimension.setdim(part2, part1.dim)

        dtype = chaospy.poly.typing.dtyping(part1.dtype, part2.dtype)

        core1 = part1.A.copy()
        core2 = part2.A.copy()

        if np.prod(part2.shape) > np.prod(part1.shape):
            shape = part2.shape
            ones = np.ones(shape, dtype=int)
            for key in core1:
                core1[key] = core1[key]*ones
        else:
            shape = part1.shape
            ones = np.ones(shape, dtype=int)
            for key in core2:
                core2[key] = core2[key]*ones

        for idx in core1:
            if idx in core2:
                core2[idx] = core2[idx] + core1[idx]
            else:
                core2[idx] = core1[idx]
        out = core2

        return Poly(out, part1.dim, shape, dtype)

    part2 = np.asarray(part2)

    core = part1.A.copy()
    dtype = chaospy.poly.typing.dtyping(part1.dtype, part2.dtype)

    zero = (0,)*part1.dim

    if zero not in core:
        core[zero] = np.zeros(part1.shape, dtype=int)

    core[zero] = core[zero] + part2

    if np.prod(part2.shape) > np.prod(part1.shape):

        ones = np.ones(part2.shape, dtype=dtype)
        for key in core:
            core[key] = core[key]*ones

    return Poly(core, part1.dim, None, dtype)


def mul(*args):
    """Polynomial multiplication."""
    if len(args) > 2:
        return add(args[0], add(args[1], args[1:]))

    if len(args) == 1:
        return args[0]

    part1, part2 = args

    if not isinstance(part2, Poly):

        if isinstance(part2, (float, int)):
            part2 = np.asarray(part2)

        if not part2.shape:
            core = part1.A.copy()
            dtype = chaospy.poly.typing.dtyping(
                part1.dtype, part2.dtype)
            for key in part1.keys:
                core[key] = np.asarray(core[key]*part2, dtype)
            return Poly(core, part1.dim, part1.shape, dtype)

        part2 = Poly(part2)

    if part2.dim > part1.dim:
        part1 = chaospy.dimension.setdim(part1, part2.dim)

    elif part2.dim < part1.dim:
        part2 = chaospy.dimension.setdim(part2, part1.dim)

    if np.prod(part1.shape) >= np.prod(part2.shape):
        shape = part1.shape
    else:
        shape = part2.shape

    dtype = chaospy.poly.typing.dtyping(part1.dtype, part2.dtype)
    if part1.dtype != part2.dtype:

        if part1.dtype == dtype:
            part2 = chaospy.poly.typing.asfloat(part2)

        else:
            part1 = chaospy.poly.typing.asfloat(part1)

    core = {}
    for idx1 in part2.A:
        for idx2 in part1.A:
            key = tuple(np.array(idx1) + np.array(idx2))
            core[key] = np.asarray(
                core.get(key, 0) + part2.A[idx1]*part1.A[idx2])
    core = {key: value for key, value in core.items() if np.any(value)}

    out = Poly(core, part1.dim, shape, dtype)
    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Create samples from the `Hammersley set`_.

The Hammersley set is equivalent to the Halton sequence, except for one
dimension is replaced with a regular grid.

Example usage
-------------

Standard usage::

    >>> print(numpy.around(create_hammersley_samples(order=3, dim=2), 4))
    [[0.75  0.125 0.625]
     [0.25  0.5   0.75 ]]
    >>> print(numpy.around(create_hammersley_samples(order=3, dim=3), 4))
    [[0.125  0.625  0.375 ]
     [0.4444 0.7778 0.2222]
     [0.25   0.5    0.75  ]]

Custom burn-ins::

    >>> print(numpy.around(create_hammersley_samples(order=3, dim=3, burnin=0), 4))
    [[0.5    0.25   0.75  ]
     [0.3333 0.6667 0.1111]
     [0.25   0.5    0.75  ]]
    >>> print(numpy.around(create_hammersley_samples(order=3, dim=3, burnin=1), 4))
    [[0.25   0.75   0.125 ]
     [0.6667 0.1111 0.4444]
     [0.25   0.5    0.75  ]]
    >>> print(numpy.around(create_hammersley_samples(order=3, dim=3, burnin=2), 4))
    [[0.75   0.125  0.625 ]
     [0.1111 0.4444 0.7778]
     [0.25   0.5    0.75  ]]

Using custom prime bases::

    >>> print(numpy.around(create_hammersley_samples(order=3, dim=2, primes=[7]), 4))
    [[0.1633 0.3061 0.449 ]
     [0.25   0.5    0.75  ]]
    >>> print(numpy.around(create_hammersley_samples(order=3, dim=3, primes=[7, 5]), 4))
    [[0.1633 0.3061 0.449 ]
     [0.64   0.84   0.08  ]
     [0.25   0.5    0.75  ]]

.. Hammersley set: https://en.wikipedia.org/wiki/Low-discrepancy_sequence#Hammersley_set
"""
import numpy

from .halton import create_halton_samples


def create_hammersley_samples(order, dim=1, burnin=None, primes=None):
    """
    Create samples from the Hammersley set.

    For ``dim == 1`` the sequence falls back to Van Der Corput sequence.

    Args:
        order (int):
            The order of the Hammersley sequence. Defines the number of samples.
        dim (int):
            The number of dimensions in the Hammersley sequence.
        burnin (int, optional):
            Skip the first ``burnin`` samples. If omitted, the maximum of
            ``primes`` is used.
        primes (array_like, optional):
            The (non-)prime base to calculate values along each axis. If
            omitted, growing prime values starting from 2 will be used.

    Returns (numpy.ndarray):
        Hammersley set with ``shape == (dim, order)``.
    """
    if dim == 1:
        return create_halton_samples(
            order=order, dim=1, burnin=burnin, primes=primes)
    out = numpy.empty((dim, order), dtype=float)
    out[:dim-1] = create_halton_samples(
        order=order, dim=dim-1, burnin=burnin, primes=primes)
    out[dim-1] = numpy.linspace(0, 1, order+2)[1:-1]
    return out



# pylint: disable=wildcard-import
"""Run through network to perform an operator."""
from .dep import dep_call
from .fwd import fwd_call
from .inv import inv_call
from .mom import mom_call
from .pdf import pdf_call
from .range_ import range_call
from .rnd import rnd_call
from .ttr import ttr_call
from .val import val_call


CALL_FUNCTIONS = {
    "dep": dep_call,
    "fwd": fwd_call,
    "inv": inv_call,
    "mom": mom_call,
    "pdf": pdf_call,
    "range": range_call,
    "rnd": rnd_call,
    "ttr": ttr_call,
    "val": val_call,
}



# pylint: disable=protected-access
"""Backend for distribution range."""
import numpy


def range_call(self, x_data, dist):
    "range call backend wrapper"

    self.counting(dist, "range")
    assert x_data.shape == (len(dist), self.size)
    graph = self.graph
    self.dist, dist_ = dist, self.dist

    graph.add_node(dist, key=x_data)
    out = numpy.empty((2,)+x_data.shape)

    prm = self.dists.build()
    prm.update(self.keys.build())
    for key, value in prm.items():
        if not isinstance(value, numpy.ndarray):
            value_ = self.run(value, "val")[0]
            if isinstance(value_, numpy.ndarray):
                prm[key] = value_
                graph.add_node(value, key=value_)

    if dist.advance:
        out[0,:],out[1,:] = dist._bnd(x_data, self)
    else:
        lower, upper = dist._bnd(**prm)
        lower, upper = numpy.array(lower), numpy.array(upper)
        out.T[:,:,0], out.T[:,:,1] = lower.T, upper.T
    graph.add_node(dist, val=out)

    self.dist = dist_
    return numpy.array(out)
